#pragma once

namespace bpftrace {

// These externs are provided by our build system. See resources/CMakeLists.txt
extern const char __stddef_max_align_t_h[];
extern const unsigned __stddef_max_align_t_h_len;
extern const char float_h[];
extern const unsigned float_h_len;
extern const char limits_h[];
extern const unsigned limits_h_len;
extern const char stdarg_h[];
extern const unsigned stdarg_h_len;
extern const char stdbool_h[];
extern const unsigned stdbool_h_len;
extern const char stddef_h[];
extern const unsigned stddef_h_len;
extern const char stdint_h[];
extern const unsigned stdint_h_len;
extern const char clang_workarounds_h[];
extern const unsigned clang_workarounds_h_len;

} // namespace bpftrace
